/* Tree Clone.
   
This file is part of GCC.
   
GCC is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.
   
GCC is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.
   
You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.  */

#include "clone.h"

static tree get_arguments (tree tree_list) {
  tree args = NULL_TREE;
  while (tree_list) {
      args = tree_cons (NULL_TREE, tree_list, args);
      tree_list = TREE_CHAIN (tree_list);
  }
  return nreverse (args);
}

static void add_call_to_clone (struct cgraph_node *orig, struct cgraph_node *clone, int feature)
{
  tree decl;
  tree and_flag, cond_expr;
  basic_block entry, call_block, body_block;
  edge_iterator ei;
  edge call_edge, body_edge, e;
  block_stmt_iterator bsi;
  struct function *func;
  struct cgraph_edge *cge;

  decl = orig->decl;
  func = DECL_STRUCT_FUNCTION (decl);

  if (DECL_SAVED_TREE (decl) != NULL)
    {
      bool removed;
      basic_block first_bb = ENTRY_BLOCK_PTR_FOR_FUNCTION (func)->next_bb;
      tree clone_stmt = build_function_call_expr (clone->decl, get_arguments(DECL_ARGUMENTS(decl)));
      tree args = get_arguments (DECL_ARGUMENTS (decl));
      tree return_stmt;

      if (TREE_TYPE (TREE_TYPE (decl)) != void_type_node) 
	{
	  tree return_var = create_tmp_var (TREE_TYPE (DECL_RESULT (decl)), "ret");
	  DECL_CONTEXT (return_var) = decl;
	  clone_stmt = build_gimple_modify_stmt (return_var, clone_stmt);
	  return_stmt = build_gimple_modify_stmt (DECL_RESULT (decl), return_var);
	  return_stmt = build1 (RETURN_EXPR, void_type_node, return_stmt);
	}
      else /* void type */
	return_stmt = build1 (RETURN_EXPR, void_type_node, NULL_TREE);

      and_flag = build2 (BIT_AND_EXPR, integer_type_node, detect_cpu_flags, build_int_cst(NULL_TREE, feature));
      cond_expr = build3 (COND_EXPR, void_type_node, and_flag, NULL_TREE, NULL_TREE);

      bsi = bsi_start (first_bb);
      #if 0
      force_gimple_operand_bsi (&bsi, cond_expr, true, NULL_TREE, true, BSI_SAME_STMT);
      #endif
      bsi_insert_before (&bsi, cond_expr, BSI_SAME_STMT);
      bsi_insert_before (&bsi, clone_stmt, BSI_SAME_STMT);

      SET_EXPR_LOCUS (clone_stmt, EXPR_LOCUS (first_stmt (first_bb)));
      TREE_BLOCK (clone_stmt) = TREE_BLOCK (first_stmt (first_bb));

      bsi_insert_before (&bsi, return_stmt, BSI_SAME_STMT);

      /* CFG corrections.  */
      bsi = bsi_start (first_bb);
      entry = bb_for_stmt (bsi_stmt (bsi));

      call_edge = split_block (entry, cond_expr);
      call_block = call_edge->dest;

      body_edge = split_block (call_block, return_stmt);
      body_block = body_edge->dest;

      /* Redirect incomming edges from entry to body_block */
      removed = true;
      while (removed)
	{
	  removed = false;
	  FOR_EACH_EDGE (e, ei, entry->preds)
	  {
	    if (e->src != ENTRY_BLOCK_PTR_FOR_FUNCTION (func))
	      {
		make_edge (e->src, body_block, e->flags);
		remove_edge (e);
		removed = true;
		break;
	      }
	  }
	}

      entry->flags |= BB_REACHABLE;
      call_block->flags |= BB_REACHABLE;
      body_block->flags |= BB_REACHABLE;

      remove_edge (body_edge);
      body_edge = make_edge (entry, body_block, EDGE_FALSE_VALUE | EDGE_EXECUTABLE);
      call_edge->flags |= EDGE_TRUE_VALUE | EDGE_EXECUTABLE;
      make_edge (call_block, EXIT_BLOCK_PTR, 0);

      body_edge->flags &= ~EDGE_FALLTHRU;
      call_edge->flags &= ~EDGE_FALLTHRU;

      entry->flags |= BB_DISABLE_SCHEDULE;
      call_block->flags |= BB_DISABLE_SCHEDULE;
      body_block->flags |= BB_DISABLE_SCHEDULE;

      update_stmt (clone_stmt);
      update_stmt (return_stmt);
      update_stmt (cond_expr);

      /* CGRAPH corrections.  */
      cge = cgraph_create_edge (orig, cgraph_node(clone->decl), clone_stmt, first_bb->count, CGRAPH_FREQ_MAX, first_bb->loop_depth);
      cge->inline_failed = "Cloned function";
      cgraph_mark_needed_node(cgraph_node(clone->decl));
    }
}

static unsigned int driver_tree_clone (void)
{
  struct cgraph_node *node;
  struct function *saved_cfun, *func;

  saved_cfun = cfun;

  for (node = cgraph_nodes; node; node = node->next) {
      func = DECL_STRUCT_FUNCTION (node->decl);
      set_cfun (func);

      if (is_clonable (node)) {
	  struct cgraph_node *clone_node = NULL;

	  if(is_renamed_function(node->decl, CLONE_PREFIX, NULL)) continue;

	  set_cfun (DECL_STRUCT_FUNCTION (node->decl));

	  clone_node = cgraph_function_versioning (node, NULL, NULL);
	  clone_node->local.inlinable = false;
	  DECL_NAME(clone_node->decl)=copy_node(DECL_NAME(node->decl));
	  rename_function(clone_node->decl, CLONE_PREFIX, "_clone");

	  if (DECL_RTL (node->decl) != NULL) {
		  SET_DECL_RTL (clone_node->decl, copy_rtx (DECL_RTL (node->decl)));
		  XEXP (DECL_RTL (clone_node->decl), 0) = gen_rtx_SYMBOL_REF (GET_MODE (XEXP (DECL_RTL (node->decl), 0)), IDENTIFIER_POINTER (DECL_NAME (clone_node->decl)));
   	  }
	
 	 //add_call_to_clone(node, clone_node, 32);
	 cgraph_mark_needed_node(clone_node);

	}

    }

  set_cfun (saved_cfun);

  return 0;
}

static bool gate_tree_clone (void)
{
  return (flag_dispatch);
}

struct tree_opt_pass pass_tree_clone = {
  "tree_clone",			/* name */
  gate_tree_clone,		/* gate */
  driver_tree_clone,		/* execute */
  NULL,				/* sub */
  NULL,				/* next */
  0,				/* static_pass_number */
  0,				/* tv_id */
  0,				/* properties_required */
  PROP_cfg,			/* properties_provided */
  0,				/* properties_destroyed */
  0,				/* todo_flags_start */
  TODO_dump_cgraph |
  TODO_dump_func |
  TODO_verify_flow |
  TODO_cleanup_cfg ,		/* todo_flags_finish */
  0				/* letter */
};

